/*
 * umaxscanner.c - driver for the Umax 1220U scanner.
 *
 * Paul Mackerras	December 1999
 *
 * Copyright 1999 Linuxcare, Inc.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include "usb.h"

struct scanner_state {
	struct usb_device *dev;
	int		epo;
	int		epi;
	struct usb_interface_descriptor *interface;
	int		present;
	int		isopen;
};

static struct scanner_state scanner;

struct ctrlmsg_ioctl {
	devrequest	req;
	void		*data;
};

static int open_scanner(struct inode *inode, struct file *file)
{
	struct scanner_state *state = &scanner;

	if (!state->present)
		return -ENODEV;

	MOD_INC_USE_COUNT;

	++state->isopen;

	file->private_data = state;
	return 0;
}

static int release_scanner(struct inode *inode, struct file *file)
{
	struct scanner_state *state = file->private_data;

	MOD_DEC_USE_COUNT;

	if (state != 0) {
		file->private_data = 0;
		--state->isopen;
	}
	return 0;
};

static int scanner_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
	struct ctrlmsg_ioctl cmsg;
	struct scanner_state *state = file->private_data;
	int pipe, ret, nb;
	struct usb_device *dev;
	unsigned char buf[64];

	if (state == 0)
		return -ENXIO;
	dev = state->dev;
	if (dev == 0)
		return -ENODEV;
	switch (cmd) {
	case 0x556d:
		if (copy_from_user(&cmsg, (void *)arg, sizeof(cmsg)))
			return -EFAULT;
		nb = le16_to_cpup(&cmsg.req.length);
		if (nb > sizeof(buf))
			return -EINVAL;
		if ((cmsg.req.requesttype & 0x80) == 0) {
			pipe = usb_sndctrlpipe(dev, 0);
			if (nb > 0 && copy_from_user(buf, cmsg.data, nb))
				return -EFAULT;
		} else
			pipe = usb_rcvctrlpipe(dev, 0);
		ret = dev->bus->op->control_msg(dev, pipe, &cmsg.req,
						buf, nb, HZ);
		if (ret < 0) {
			printk(KERN_DEBUG "control_msg returned %d\n", ret);
			return -EIO;
		}
		if (nb > 0 && (cmsg.req.requesttype & 0x80)
		    && copy_to_user(cmsg.data, buf, nb))
			return -EFAULT;
		return 0;

	default:
		return -ENOTTY;
	}
			
}

static ssize_t write_scanner(struct file *file, const char *buffer,
			     size_t count, loff_t *ppos)
{
	struct scanner_state *state = file->private_data;
	int ret;
	ssize_t err;
	struct usb_device *dev;
	char *buf;
	unsigned long rval;

	if (state == 0)
		return -ENXIO;
	dev = state->dev;
	if (dev == 0)
		return -ENXIO;
	if (count > 0xffff)
		return -EINVAL;
	buf = kmalloc(count, GFP_KERNEL);
	if (buf == 0)
		return -ENOMEM;
	err = -EFAULT;
	if (copy_from_user(buf, buffer, count))
		goto outf;

	err = -EIO;
	ret = usb_control_msg(dev, usb_sndctrlpipe(dev, 0), 3, 0x40,
			      count, 0, 0, 0, HZ);
	if (ret < 0) {
		printk(KERN_INFO "scanner write msg returned %d\n", ret);
		goto outf;
	}

	ret = usb_bulk_msg(dev, usb_sndbulkpipe(dev, state->epo),
			   buf, count, &rval, HZ*10);
	if (ret < 0) {
		printk(KERN_INFO "scanner write bulk returned %d, rval %ld\n",
		       ret, rval);
		goto outf;
	}
	if (rval != count) {
		printk(KERN_INFO "scanner write rval=%ld count=%ld pipe=%x\n",
		       rval, count, usb_sndbulkpipe(dev, state->epo));
	}
	file->f_dentry->d_inode->i_atime = CURRENT_TIME;

	err = count;

 outf:
	kfree(buf);
	return err;
}

static ssize_t read_scanner(struct file *file, char *buffer, size_t count,
			    loff_t *ppos)
{
	struct scanner_state *state = file->private_data;
	int ret;
	ssize_t err;
	struct usb_device *dev;
	char *buf;
	unsigned long rval;

	if (state == 0)
		return -ENXIO;
	dev = state->dev;
	if (dev == 0)
		return -ENXIO;
	if (count > 0xffff)
		return -EINVAL;
	buf = kmalloc(count, GFP_KERNEL);
	if (buf == 0)
		return -ENOMEM;

	err = -EIO;
	ret = usb_control_msg(dev, usb_sndctrlpipe(dev, 0), 2, 0x40,
			      count, 0, 0, 0, HZ);
	if (ret < 0) {
		printk(KERN_INFO "scanner read msg returned %d\n", ret);
		goto outf;
	}

	ret = usb_bulk_msg(dev, usb_rcvbulkpipe(dev, state->epi),
			   buf, count, &rval, HZ*5);
	if (ret == USB_ST_TIMEOUT && rval != 0)
		count = rval;
	else if (ret < 0 || rval != count) {
		printk(KERN_INFO "scanner read bulk returned %d, rval %ld\n",
		       ret, rval);
		goto outf;
	}
	file->f_dentry->d_inode->i_atime = CURRENT_TIME;

	err = -EFAULT;
	if (copy_to_user(buffer, buf, count))
		goto outf;
	err = count;

 outf:
	kfree(buf);
	return err;
}

struct file_operations umaxscanner_fops = {
	NULL,		/* seek */
	read_scanner,
	write_scanner,
	NULL,		/* readdir */
	NULL,		/* poll */
	scanner_ioctl,
	NULL,		/* mmap */
	open_scanner,
	NULL,		/* flush */
	release_scanner,
	NULL,
	NULL		/* fasync */
};

static void *umaxscanner_probe(struct usb_device *dev, unsigned int i)
{
	struct usb_interface_descriptor *interface;
	struct scanner_state *state;

	if (dev->descriptor.idVendor != 0x1606
	    || dev->descriptor.idProduct != 0x10)
		return NULL;
	interface = &dev->actconfig->interface[i].altsetting[0];
	if (interface->bNumEndpoints != 3)
		return NULL;
	/* XXX should analyse the endpoints here */

	state = &scanner;

	state->dev = dev;
	state->epo = 1;
	state->epi = 2;
	state->interface = interface;
	state->present = 1;

	printk(KERN_INFO "USB scanner found\n");

	return state;
}

static void umaxscanner_disconnect(struct usb_device *dev, void *ptr)
{
	struct scanner_state *state = ptr;

	state->dev = 0;
	state->present = 0;
	printk(KERN_INFO "USB scanner disconnected\n");
}

static struct usb_driver umaxscanner_driver = {
	"umaxscanner",
	umaxscanner_probe,
	umaxscanner_disconnect,
	{ NULL, NULL },
	&umaxscanner_fops,
	192
};

int umaxscanner_init(void)
{
	if (usb_register(&umaxscanner_driver) < 0)
		return -1;
	return 0;
}

void umaxscanner_cleanup(void)
{
	usb_deregister(&umaxscanner_driver);
}

#ifdef MODULE
int init_module(void)
{
	return umaxscanner_init();
}

void cleanup_module(void)
{
	umaxscanner_cleanup();
}
#endif
